% Generated by roxygen2 (4.1.1): do not edit by hand
% Please edit documentation in R/DistToNearest.R
\name{distToNearest}
\alias{distToNearest}
\title{Distance to nearest neighbor}
\usage{
distToNearest(db, sequenceColumn = "JUNCTION", vCallColumn = "V_CALL",
  jCallColumn = "J_CALL", model = c("hs1f", "m1n", "ham", "aa", "hs5f"),
  normalize = c("length", "none"), first = TRUE, nproc = 1)
}
\arguments{
\item{db}{data.frame containing sequence data.}

\item{sequenceColumn}{name of the column containing nucleotide sequences to compare.
Also used to determine sequence length for grouping.}

\item{vCallColumn}{name of the column containing the V-segment allele calls.}

\item{jCallColumn}{name of the column containing the J-segment allele calls.}

\item{model}{underlying SHM model, which must be one of
\code{c("m1n", "ham", "aa", "hs5f")}.
See Details for further information.}

\item{normalize}{method of normalization. The default is "none". If the "length"
method is chosen, then distance is divided by the length of the
junction sequence.}

\item{first}{if \code{TRUE} only the first call of the gene assignments is used.
If \code{FALSE} the union of ambiguous gene assignments is used to
group all sequences with any overlapping gene calls.}

\item{nproc}{number of cores to distribute the function over.}
}
\value{
Returns a modified \code{db} data.frame with nearest neighbor distances in the
          \code{DIST_NEAREST} column.
}
\description{
Get distance of every sequence to its nearest sequence sharing same V gene, J gene, and
sequence length.
}
\details{
The distance to nearest neighbor can be used to estimate a threshold for assigning Ig
sequences to clonal groups. A histogram of the resulting vector is often bimodal,
with the ideal threshold being a value that separates the two modes.

"hs5f" use distance derived from the \link{HS5FModel}
using \link{calcTargetingDistance}. "hs1f" and "m1n" use \link{HS1FDistance} and \link{M1NDistance}
 to calculate distances respectively. "ham" uses a nucleotide hamming distance matrix from
 \link{getDNADistMatrix}, with gaps being zero. "aa" uses an amino acid hamming distance matrix
 from \link{getAADistMatrix}.
}
\examples{
# Load example data
library(alakazam)
file <- system.file("extdata", "changeo_demo.tab", package="alakazam")
db <- readChangeoDb(file)

# Use genotyped V assignments and HS5F model
dist_hs5f <- distToNearest(db, vCallColumn="V_CALL_GENOTYPED", model="hs5f", first=FALSE)
hist(dist_hs5f$DIST_NEAREST, breaks=100, xlim=c(0, 60))

# Use HS1F model and normalize by junction length
dist_hs1f <- distToNearest(db, model="hs1f", first=FALSE, normalize="length")
hist(dist_hs1f$DIST_NEAREST, breaks=25, xlim=c(0, 1))
}
\references{
\enumerate{
  \item  Smith DS, et al. Di- and trinucleotide target preferences of somatic
           mutagenesis in normal and autoreactive B cells.
           J Immunol. 1996 156:2642-52.
  \item  Glanville J, Kuo TC, von Budingen H-C, et al.
           Naive antibody gene-segment frequencies are heritable and unaltered by
           chronic lymphocyte ablation.
           Proc Natl Acad Sci USA. 2011 108(50):20066-71.
  \item  Yaari G, et al. Models of somatic hypermutation targeting and substitution based
           on synonymous mutations from high-throughput immunoglobulin sequencing data.
           Front Immunol. 2013 4:358.
 }
}
\seealso{
See \link{calcTargetingDistance} for generating nucleotide distance matrices
          from a \link{TargetingModel} object. See \link{M1NDistance},
          \link{HS5FModel}, \link{getDNADistMatrix}, and \link{getAADistMatrix}
          for individual model details.
}

